#ifndef IOCLASS_HH
#define IOCLASS_HH

#include "eventwrap.hh"
#include "sockaddr.hh"

class io_class : public event
{
protected:
	int fd;
	bool own;
public:
	io_class(int _fd, bool _own = true);
	virtual ~io_class();

	int fileno() const { return fd; }
protected:
	virtual void poll_read() {}
	virtual void poll_write() {}
	virtual void poll_error(int err) { throw int(err); }
	virtual void poll_timedout() {}

	void handle(int _fd, int evflags);
};

class io_socket : public io_class
{
public:
	io_socket(int fd, bool own = true);
	io_socket(int domain, int type, int proto = 0);

	void bind(const socket_address& sa);
	void get_my_address(socket_address& sa) const;
	void get_peer_address(socket_address& sa) const;

	socket_address get_my_address() const { socket_address sa; get_my_address(sa); return sa; }
	socket_address get_peer_address() const { socket_address sa; get_peer_address(sa); return sa; }
};

class io_stream_listener : public io_socket
{
protected:
	bool listening;
public:
	io_stream_listener(int domain, int proto = 0);
	~io_stream_listener();

	void listen();
protected:
	virtual void poll_newconn(int fd) = 0;

	void poll_read();
};

class io_stream_connector : public io_socket
{
public:
	static timeout default_connect_timeout;

	io_stream_connector(int domain, int proto = 0);

	void connect(const socket_address& sa, const timeout& to = default_connect_timeout);
protected:
	virtual void poll_connected() = 0;

	void poll_write();
};

class address_detector : public io_stream_connector
{
public:
	address_detector(int domain, int proto = 0)
		: io_stream_connector(domain, proto) {}

	void execute(const socket_address& to, socket_address& dest);
protected:
	void poll_connected() {}
};

class io_datagram_socket : public io_socket
{
public:
	io_datagram_socket(int domain, int proto = 0);

	void associate(const socket_address& peer);
};

#endif /* !IOCLASS_HH */
