#include <errno.h>
#include <iostream>
using namespace std;
#include "packetio.hh"
#include "logger.hh"

void packet_io::wait_for_write(bool wait)
{
	del(*this);
	set(wait ? EV_READ|EV_WRITE|EV_PERSIST : EV_READ|EV_PERSIST);
	add(*this);
}

void packet_io::poll_read()
{
	ssize_t n = ::recv(fd, NULL, 0, MSG_PEEK|MSG_TRUNC);

	if (n <= 0) {
		if (n == 0)
			poll_error(0);
		else if (errno != EINTR && errno != EAGAIN)
			poll_error(errno);
		return;
	}

	ssize_t e = ::recv(fd, packet.append_space(n), n, 0);

	if (n != e) {
		if (DEBUG(PACKETFLOW))
			logger	<< timestamp << " error while recving packet on [" << fd << "]: "
				<< errno_str(n < 0 ? errno : EIO) << endl;
		poll_error(e < 0 ? errno : EIO);
		return;
	}

	if (DEBUG(PACKETDATA))
		logger << timestamp << " packet to [" << fd << "]:" << packet;
	else if (DEBUG(PACKETFLOW))
		logger << timestamp << " packet to [" << fd << "] len " << packet.count_elements_left() << endl;

	pkt_read();
	packet.clear();
}

void packet_io::poll_write()
{
	if (out_queue.is_empty()) {
		logger << timestamp << " **BUG** poll_write() with empty queue on [" << fd << ']' << endl;
		finish_write();
		return;
	}

	virtual_buffer<char> pkt = out_queue.get_next();

	ssize_t n = ::send(fd, pkt.get_head_ptr(), pkt.count_elements_left(), MSG_DONTWAIT|MSG_NOSIGNAL);

	if (n <= 0) {
		if (DEBUG(PACKETFLOW))
			logger	<< timestamp << " error while sending packet on [" << fd << "]: "
				<< errno_str(n == 0 ? 0 : errno) << endl;
		if (n == 0)
			poll_error(0);
		else if (errno != EINTR && errno != EAGAIN)
			poll_error(errno);
	} else {
		if (DEBUG(PACKETDATA))
			logger	<< timestamp << " packet sent to [" << fd << "]:" << pkt;
		else if (DEBUG(PACKETFLOW))
			logger	<< timestamp << " packet sent on [" << fd << "] len " << pkt.count_elements_left() << endl;
		out_queue.pop();
		finish_write();
	}
}

void addressed_packet_io::wait_for_write(bool wait)
{
	del(*this);
	set(wait ? EV_READ|EV_WRITE|EV_PERSIST : EV_READ|EV_PERSIST);
	add(*this);
}

void addressed_packet_io::poll_read()
{
	ssize_t n = ::recv(fd, NULL, 0, MSG_PEEK|MSG_TRUNC);

	if (n <= 0) {
		if (DEBUG(PACKETFLOW))
			logger	<< timestamp << " error while checking packet on [" << fd << "]: "
				<< errno_str(n == 0 ? 0 : errno) << endl;
		if (n == 0)
			poll_error(0);
		else if (errno != EINTR && errno != EAGAIN)
			poll_error(errno);
		return;
	}

	packet_from.salen = packet_from.salen_max();
	ssize_t e = ::recvfrom(fd, packet.append_space(n), n, 0, packet_from.sockaddr(), &packet_from.salen);

	if (n != e) {
		if (DEBUG(PACKETFLOW))
			logger	<< timestamp << " error while recving packet on [" << fd << "]: "
				<< errno_str(n < 0 ? errno : EIO) << endl;
		poll_error(e < 0 ? errno : EIO);
		return;
	}

	if (DEBUG(PACKETDATA))
		logger	<< timestamp << " packet to [" << fd << "] from " << packet_from << ':' << packet;
	else if (DEBUG(PACKETFLOW))
		logger	<< timestamp << " packet to [" << fd << "] from " << packet_from
			<< " len " << packet.count_elements_left() << endl;

	pkt_read();
	packet.clear();
}

void addressed_packet_io::poll_write()
{
	if (out_queue.is_empty()) {
		logger << timestamp << " **BUG** poll_write() with empty queue on [" << fd << ']' << endl;
		finish_write();
		return;
	}

	virtual_buffer<char> pkt = out_queue.get_next();
	socket_address to = addr_q.front();

	ssize_t n = ::sendto(fd, pkt.get_head_ptr(), pkt.count_elements_left(), MSG_DONTWAIT,
		to.const_sockaddr(), to.salen);

	if (n <= 0) {
		if (DEBUG(PACKETFLOW))
			logger	<< timestamp << " error while sending packet on [" << fd << "] to " << to
				<< ": " << errno_str(n == 0 ? 0 : errno) << endl;
		if (n == 0)
			poll_error(0);
		else if (errno != EINTR && errno != EAGAIN)
			poll_error(errno);
	} else {
		if (DEBUG(PACKETFLOW))
			logger	<< timestamp << " packet sent on [" << fd << "] to " << to
				<< " len " << pkt.count_elements_left() << endl;
		addr_q.pop();
		out_queue.pop();
		finish_write();
	}
}


