#ifndef PACKETIO_HH
#define PACKETIO_HH

#include "ioclass.hh"
#include "buffer.hh"
#include <sys/types.h>

class pkt_io_buffer
{
public:
	buffer<char> packet;
protected:
	packet_buffer<char> out_queue;
	bool waiting_for_write;

	virtual void wait_for_write(bool wait) = 0;

	void start_write() {
		if (!out_queue.is_empty() && !waiting_for_write)
			wait_for_write(waiting_for_write = true);
	}
	void finish_write() {
		if (waiting_for_write && out_queue.is_empty())
			wait_for_write(waiting_for_write = false);
	}

	virtual void pkt_read() = 0;
public:
	pkt_io_buffer() : waiting_for_write(false) {}
	virtual ~pkt_io_buffer() {}

	void cleanbufs() { out_queue.clean(); }

	void out_add(const virtual_buffer<char>& buf)
		{ out_queue.push(buf); start_write(); }
};

class addressed_pkt_io_buffer : public pkt_io_buffer
{
public:
	socket_address packet_from;
protected:
	std::queue<socket_address> addr_q;
public:
	addressed_pkt_io_buffer() : pkt_io_buffer() {}

	void cleanbufs() { pkt_io_buffer::cleanbufs(); }

	void out_add_to(const virtual_buffer<char>& buf, const socket_address& dest)
		{ addr_q.push(dest); out_add(buf); }
};

class packet_io : public io_datagram_socket, public pkt_io_buffer
{
protected:
	void wait_for_write(bool wait);
	void poll_read();
	void poll_write();
public:
	packet_io(int _fd, bool _own) : io_datagram_socket(_fd, _own) {}
	packet_io(int domain, int proto = 0) : io_datagram_socket(domain, proto) {}
};

class addressed_packet_io : public io_datagram_socket, public addressed_pkt_io_buffer
{
protected:
	void wait_for_write(bool wait);
	void poll_read();
	void poll_write();
public:
	addressed_packet_io(int _fd, bool _own) : io_datagram_socket(_fd, _own) {}
	addressed_packet_io(int domain, int proto = 0) : io_datagram_socket(domain, proto) {}
};

#endif /* !PACKETIO_HH */
