#ifndef TCPFORWARD_HH
#define TCPFORWARD_HH

#include "eventwrap.hh"
#include "ioclass.hh"
#include "bufdio.hh"
#include <set>

class tcp_forward
{
protected:
	class forwarding_side : public buffered_io
	{
		friend class tcp_forward;
	protected:
		tcp_forward *master;
	public:
		forwarding_side(tcp_forward *_master, int _fd)
			: buffered_io(_fd), master(_master) {}

		void start();
		void stop();
	protected:
		void buf_read();
		void poll_error(int err);
	};

	friend class forwarding_side;
	enum forward_dir { TO_CLIENT = 0, TO_SERVER = 1 };

	forwarding_side *slave[2];

	bool is_client_slave(const forwarding_side *s) const { return s == slave[TO_CLIENT]; }
	buffered_io *get_client_io() { return slave[TO_CLIENT]; }
	bool is_server_slave(const forwarding_side *s) const { return s == slave[TO_SERVER]; }
	buffered_io *get_server_io() { return slave[TO_SERVER]; }

	virtual void slave_read_callback(forwarding_side *_slave);
	virtual void slave_error_callback(forwarding_side *_slave, int err);
public:
	tcp_forward() { slave[0] = slave[1] = NULL; }
	~tcp_forward();

	void start_forward(int fds[2]);
	int fileno(int i) const { return slave[i]->fileno(); }
};

class tcp_forward_listener : public io_stream_listener
{
protected:
	class forward_creator : public io_stream_connector
	{
	private:
		int sibling_fd;

		tcp_forward_listener *master;
	public:
		forward_creator(tcp_forward_listener *_master, int sfd)
			: io_stream_connector(PF_INET), sibling_fd(sfd), master(_master) {}
		~forward_creator() { if (own) ::close(sibling_fd); }

		int sibling_fileno() const { return sibling_fd; }
	protected:
		void poll_connected();
		void poll_error(int err);
	};

	friend class forward_creator;
	typedef std::set<forward_creator *> newconn_set;

	newconn_set newconns;
	socket_address to;

	virtual tcp_forward *create_forward();
	virtual void connected(tcp_forward *fwd) {}
	virtual void connect_died(int ofd, int cfd, int err) {}
	virtual bool validate_connection(int ofd, int cfd) { return true; }

	void poll_newconn(int fd);
private:
	void start_forwarding(forward_creator *fwdc, tcp_forward *fwd);
	void forward_died(forward_creator *fwdc, int err);
public:
	tcp_forward_listener(const socket_address& _to) : io_stream_listener(PF_INET), to(_to) {}
	~tcp_forward_listener();
};

class tcp_proxied_connection : public tcp_forward
{
protected:
	virtual void read_from_client() = 0;
	virtual void read_from_server() = 0;
	virtual void connection_error(enum forward_dir dir, int err);

	void slave_read_callback(forwarding_side *_slave);
	void slave_error_callback(forwarding_side *_slave, int err);
public:
	void send_to_client(const virtual_buffer<char>& buf) { slave[0]->out_append(buf); }
	void send_to_server(const virtual_buffer<char>& buf) { slave[1]->out_append(buf); }
};

#endif /* !TCPFORWARD_HH */
