#include <sys/types.h>
#include <sys/socket.h>
#include <iostream>
using namespace std;
#include "udpmx.hh"
#include "logger.hh"

udp_mx::udp_mx()
	: addressed_packet_io(PF_INET), running(false)
{
}

udp_mx::~udp_mx()
{
	for(peermap::iterator p = peers.begin(); p != peers.end(); ++p)
		delete p->second;

	if (running && DEBUG(PACKETFLOW))
		logger	<< timestamp << " UDP mx [" << fd << "] closed" << endl;
}

void udp_mx::pkt_read()
{
	peermap::iterator p = peers.find(packet_from);

	if (p == peers.end()) {
		bool b = validate_new_peer();
		if (DEBUG(PACKETFLOW))
			logger	<< timestamp << " new peer on [" << fd << "] from " << packet_from
				<< (b ? "" : " NOT") << " accepted" << endl;
		if (!b)
			return;

		p = peers.insert(peermap::value_type(packet_from, create_peer())).first;
	}
	p->second->pkt_read();
	packet.clear();
}

void udp_mx::start()
{
	event::set(EV_READ|EV_PERSIST);
	event::add(*this);
	running = true;

	if (DEBUG(PACKETFLOW)) {
		socket_address me;
		get_my_address(me);
		logger	<< timestamp << " UDP mx [" << fd << "] listening on " << me << endl;
	}
}

void udp_mx::add_peer(udp_mx_peer *peer)
{
	if (DEBUG(PACKETFLOW))
		logger	<< timestamp << " new peer on [" << fd << "] from " << peer->remote
			<< " forced" << endl;
	peers.insert(peermap::value_type(peer->remote, peer));
}

void udp_mx::drop_peer(udp_mx_peer *peer)
{
	if (DEBUG(PACKETFLOW))
		logger	<< timestamp << " dropped peer on [" << fd << "] from " << peer->remote << endl;
	peers.erase(peer->remote);
	delete peer;
}

udp_mx_peer *udp_mx::find_peer(const socket_address::ipportpair& addr)
{
	peermap::iterator p = peers.find(addr);
	if (p == peers.end())
		return NULL;

	return p->second;
}

